use tokio::{
	net::TcpListener,
};

use quic_demo::prelude::*;

#[tokio::main]
async fn main () -> anyhow::Result <()> {
	tracing_subscriber::fmt::init ();
	
	let server_cert = tokio::fs::read ("quic_server.crt").await?;
	let server_addr = "127.0.0.1:30380".parse ()?;
	let endpoint = make_client_endpoint ("0.0.0.0:0".parse ()?, &[&server_cert])?;
	
	debug! ("Connecting to relay server");
	
	let quinn::NewConnection {
		connection,
		..
	} = endpoint.connect (&server_addr, "localhost")?.await?;
	
	let (mut send, mut recv) = connection.open_bi ().await?;
	
	let req_buf = [2u8, 42, 0, 0];
	send.write_all (&req_buf).await?;
	
	let mut resp_buf = [0u8, 0, 0, 0];
	recv.read_exact (&mut resp_buf).await?;
	
	debug! ("Waiting for local TCP client to connect to us");
	
	let listener = TcpListener::bind ("127.0.0.1:30381").await?;
	let (tcp_socket, _) = listener.accept ().await?;
	let (local_recv, local_send) = tcp_socket.into_split ();
	
	debug! ("Connecting to end server");
	
	let (mut relay_send, mut relay_recv) = connection.open_bi ().await?;
	
	let req_buf = [1, 43, 0, 0, 1, 0, 0, 0];
	relay_send.write_all (&req_buf).await?;
	
	let mut resp_buf = [0; 8];
	relay_recv.read_exact (&mut resp_buf).await?;
	
	debug! ("Relaying bytes...");
	
	let ptth_conn = PtthNewConnection {
		local_send,
		local_recv,
		relay_send,
		relay_recv,
	}.build ();
	
	ptth_conn.uplink_task.await??;
	ptth_conn.downlink_task.await??;
	
	Ok (())
}

struct PtthNewConnection {
	local_send: tokio::net::tcp::OwnedWriteHalf,
	local_recv: tokio::net::tcp::OwnedReadHalf,
	relay_send: quinn::SendStream,
	relay_recv: quinn::RecvStream,
}

struct PtthConnection {
	uplink_task: JoinHandle <anyhow::Result <()>>,
	downlink_task: JoinHandle <anyhow::Result <()>>,
}

impl PtthNewConnection {
	fn build (self) -> PtthConnection {
		let Self {
			mut local_send,
			mut local_recv,
			mut relay_send,
			mut relay_recv,
		} = self;
		
		let uplink_task = tokio::spawn (async move {
			// Uplink - local client to relay server
			
			let mut buf = vec! [0u8; 65_536];
			loop {
				let bytes_read = local_recv.read (&mut buf).await?;
				if bytes_read == 0 {
					break;
				}
				let buf_slice = &buf [0..bytes_read];
				debug! ("Uplink relaying {} bytes", bytes_read);
				relay_send.write_all (buf_slice).await?;
			}
			
			debug! ("Uplink closed");
			
			Ok::<_, anyhow::Error> (())
		});

		let downlink_task = tokio::spawn (async move {
			// Downlink - Relay server to local client
			
			let mut buf = vec! [0u8; 65_536];
			while let Some (bytes_read) = relay_recv.read (&mut buf).await? {
				let buf_slice = &buf [0..bytes_read];
				debug! ("Downlink relaying {} bytes", bytes_read);
				local_send.write_all (buf_slice).await?;
			}
			
			debug! ("Downlink closed");
			
			Ok::<_, anyhow::Error> (())
		});
		
		PtthConnection {
			uplink_task,
			downlink_task,
		}
	}
}
