use tokio::net::TcpStream;

use quic_demo::prelude::*;

#[tokio::main]
async fn main () -> anyhow::Result <()> {
	tracing_subscriber::fmt::init ();
	
	let server_cert = tokio::fs::read ("quic_server.crt").await?;
	let server_addr = "127.0.0.1:30380".parse ()?;
	let endpoint = make_client_endpoint ("0.0.0.0:0".parse ()?, &[&server_cert])?;
	
	debug! ("Connecting to relay server");
	
	let quinn::NewConnection {
		connection,
		mut bi_streams,
		..
	} = endpoint.connect (&server_addr, "localhost")?.await?;
	
	let (mut send, mut recv) = connection.open_bi ().await?;
	
	let req_buf = [4u8, 43, 0, 0];
	send.write_all (&req_buf).await?;
	
	let mut resp_buf = [0u8, 0, 0, 0];
	recv.read_exact (&mut resp_buf).await?;
	
	debug! ("Waiting for relay server to forward a bi stream");
	
	let (mut relay_send, mut relay_recv) = bi_streams.next ().await.ok_or_else (|| anyhow::anyhow! ("Relay server didn't open a bi stream"))??;
	
	let mut req_buf = [0, 0, 0, 0];
	relay_recv.read_exact (&mut req_buf).await?;
	
	// TODO: Auth stuff
	
	debug! ("Connecting to TCP end server");
	
	let stream = TcpStream::connect ("127.0.0.1:30382").await?;
	let (mut local_recv, mut local_send) = stream.into_split ();
	
	let resp_buf = [20, 0, 0, 0];
	relay_send.write_all (&resp_buf).await?;
	relay_send.write_all (&resp_buf).await?;
	
	debug! ("Relaying bytes...");
	
	let uplink_task = tokio::spawn (async move {
		// Downlink - Relay server to local client
		
		let mut buf = vec! [0u8; 65_536];
		while let Some (bytes_read) = relay_recv.read (&mut buf).await? {
			let buf_slice = &buf [0..bytes_read];
			debug! ("Uplink relaying {} bytes", bytes_read);
			local_send.write_all (buf_slice).await?;
		}
		
		debug! ("Downlink closed");
		
		Ok::<_, anyhow::Error> (())
	});
	
	let downlink_task = tokio::spawn (async move {
		// Uplink - local client to relay server
		
		let mut buf = vec! [0u8; 65_536];
		loop {
			let bytes_read = local_recv.read (&mut buf).await?;
			if bytes_read == 0 {
				break;
			}
			let buf_slice = &buf [0..bytes_read];
			debug! ("Downlink relaying {} bytes", bytes_read);
			relay_send.write_all (buf_slice).await?;
		}
		
		debug! ("Uplink closed");
		
		Ok::<_, anyhow::Error> (())
	});
	
	uplink_task.await??;
	downlink_task.await??;
	
	Ok (())
}
